unit app_cmd;

interface

uses Windows, Messages, SysUtils, Classes, ExtCtrls, Math, OverbyteIcsConApp,
     app_utils, obimp_dll, u_obimp_const, u_obimp_codes;

type
  TFuncCreateObimp = function(ObimpEvents: IObimpEvents): IObimpProto; stdcall;

  TObimpEvents = class(TInterfacedObject, IObimpEvents)
  public
    {==== Interface functions ====}
    function  GetProtoIntfVer(): DWord; stdcall;
    procedure ConnectionFailed(); stdcall;
    procedure LoginSuccess(); stdcall;
    procedure LoginFailed(LoginFailedCode: Word); stdcall;
    procedure ServerBye(ByeReasonCode: Word); stdcall;
    procedure MsgReceived(RcvdMsgInfo: pRcvdMsgInfo); stdcall;
    procedure MsgAckRcvd(MsgAck: pMsgAck); stdcall;
    procedure AuthReqRcvd(AuthReq: pRcvdAuthReq); stdcall;
    procedure AuthReplyRcvd(AuthReply: pRcvdAuthReply); stdcall;
    procedure AuthRevokeRcvd(AuthRevoke: pRcvdAuthRevoke); stdcall;
  end;

  TBimAppCmd = class(TConApplication)
  private
    FLibHandle      : THandle;
    FuncCreateObimp : TFuncCreateObimp;
    FUniqSeq        : DWord;
    FConnected      : Boolean;
    FProtoEvents    : TObimpEvents;
    FCmdParams      : TAppCmdParams;
    FListSend       : TStringList;
    FTimerSend      : TTimer;
    FTimerRecon     : TTimer;
    FSendCount      : Cardinal;
    procedure QuitApp;
    function  GetUniqSequence: DWord;
    procedure SendMsgToAcc(Acc: string);
    procedure OnAppCmdSendTimer(Sender: TObject);
    procedure OnAppCmdReconTimer(Sender: TObject);
  public
    ObimpEvents : IObimpEvents;
    ObimpProto  : IObimpProto;
    constructor Create(Parent:TComponent); override;
    destructor  Destroy; override;
    procedure   DoCharReceived(Ch : Char); override;
    procedure   Execute; override;
    procedure   Start;
    function    TryToLogin: Boolean;
    procedure   ReconnectRequired;
    procedure   ReadyToSendMsg;
  end;

var
  AppCmd: TBimAppCmd;

implementation

{ TBimAppCmd }
{***************************************************************}
constructor TBimAppCmd.Create(Parent:TComponent);
begin
  //Create protocol events class
  FProtoEvents := TObimpEvents.Create;

  //Increase interface references counter, else our events class will be
  //freed after destroying ObimpProto interface.
  ObimpEvents := FProtoEvents;

  //Dynamically loading micro OBIMP library...
  FLibHandle := SafeLoadLibrary(OBIMP_DLL_FILE_NAME);
  if (FLibHandle = 0) then
  begin
    Log('* "' + OBIMP_DLL_FILE_NAME + '" file not found! App will be terminated!');
    Halt;
  end
  else
    Log('* Library "' + OBIMP_DLL_FILE_NAME + '" loaded successfully.');

  //Getting CreateObimp function address from OBIMP library...
  @FuncCreateObimp := GetProcAddress(FLibHandle, DLL_EXP_FNAME_CREATE_OBIMP);
  if (@FuncCreateObimp = nil) then
  begin
    Log('* Function "' + DLL_EXP_FNAME_CREATE_OBIMP + '" not found! App will be terminated!');
    FreeLibrary(FLibHandle);
    Halt;
  end;

  //Creating OBIMP protocol interface instance
  ObimpProto := FuncCreateObimp(ObimpEvents);

  //Interface version can be used for compatibility purposes
  Log('* ObimpProto interface version ' + IntToStr(ObimpProto.GetProtoIntfVer));

  FTimerSend := TTimer.Create(nil);
  FTimerSend.Enabled := False;
  FTimerSend.OnTimer := OnAppCmdSendTimer;

  FTimerRecon := TTimer.Create(nil);
  FTimerRecon.Enabled := False;
  FTimerRecon.OnTimer := OnAppCmdReconTimer;

  FListSend := TStringList.Create;
end;

{***************************************************************}
destructor TBimAppCmd.Destroy;
begin
  if Assigned(FTimerRecon) then
    FreeAndNil(FTimerRecon);

  if Assigned(FTimerSend) then
    FreeAndNil(FTimerSend);

  if Assigned(FListSend) then
    FreeAndNil(FListSend);

  //destroy protocol interface instance
  ObimpProto := nil;

  //nil events interface, it will free FProtoEvents class
  ObimpEvents := nil;

  if (FLibHandle > 0) then
    FreeLibrary(FLibHandle);

  inherited;
end;

{***************************************************************}
procedure TBimAppCmd.DoCharReceived(Ch: Char);
begin
  // Test for ESC key
  if (Ch = #27) then
    QuitApp;
end;

{***************************************************************}
procedure TBimAppCmd.Execute;
begin
  // Very simple here: just call the message pump until terminated
  while not Terminated do
    MessageLoop;
end;

{***************************************************************}
function TBimAppCmd.GetUniqSequence: DWord;
begin
  //unique sequence needed for delivery message acknowledges
  Inc(FUniqSeq);
  Result := FUniqSeq;
end;

{***************************************************************}
procedure TBimAppCmd.QuitApp;
begin
  Terminate;
end;

{***************************************************************}
procedure TBimAppCmd.Start;
var bFailed: Boolean;
    i: Integer;
begin
  bFailed := True;
  try
    //get command line params
    GetCmdParams(FCmdParams);

    //show help if specified
    if FCmdParams.OtherInfo.ShowHelp then
    begin
      Log(HELP_INFO);
      Exit;
    end;

    if (FCmdParams.MsgInfo.SendTo = '') then
    begin
      Log('! No recipients specified.');
      if not FCmdParams.OtherInfo.StayConnected then
        Exit;
    end;

    if (FCmdParams.MsgInfo.MsgText = '') then
    begin
      //clear recipients too
      FCmdParams.MsgInfo.SendTo := '';

      Log('! No message text specified.');
      if not FCmdParams.OtherInfo.StayConnected then
        Exit;
    end;

    FListSend.Text := FCmdParams.MsgInfo.SendTo;

    //clean empty lines if available
    for i := FListSend.Count-1 downto 0 do
    begin
      if (Trim(FListSend.Strings[i]) = '') then
        FListSend.Delete(i);
    end;//for

    if (FCmdParams.OtherInfo.Interval < 250) then
      FCmdParams.OtherInfo.Interval := 250
    else
    if (FCmdParams.OtherInfo.Interval > 60000) then
      FCmdParams.OtherInfo.Interval := 60000;

    Log('* Message sending interval: ' + IntToStr(FCmdParams.OtherInfo.Interval) + ' msec');

    bFailed := not TryToLogin;
  finally
    //terminate exe if failed
    if bFailed then
      QuitApp;
  end;
end;

{***************************************************************}
function TBimAppCmd.TryToLogin: Boolean;
var iRes: integer;
begin
  Result := False;

  //connect to server
  iRes := ObimpProto.Login(@FCmdParams.LoginInfo, @FCmdParams.ClientInfo, @FCmdParams.StatusInfo);
  if (iRes = FUNC_OPER_SUCCESS) then
  begin
    Log('* Connecting ...');
    Result := True;
  end
  else
    Log('! Can''t login. Some parameter is wrong.');
end;

{***************************************************************}
procedure TBimAppCmd.ReadyToSendMsg;
begin
  if FCmdParams.OtherInfo.StayConnected then
    Log('* Staying connected');

  FTimerSend.Interval := FCmdParams.OtherInfo.Interval;
  FTimerSend.Enabled := True;
end;

{***************************************************************}
procedure TBimAppCmd.ReconnectRequired;
begin
  if not FCmdParams.OtherInfo.NoReconnect then
  begin
    FTimerSend.Enabled := False;

    FTimerRecon.Interval := FTimerRecon.Interval + Cardinal(IfThen(FTimerRecon.Interval = 1000, 4000, 5000));
    if (FTimerRecon.Interval > 60000) then
      FTimerRecon.Interval := 1000;

    Log('* Going to reconnect, interval=' + IntToStr(FTimerRecon.Interval div 1000) + 's');

    FTimerRecon.Enabled := True;
  end
  else
    QuitApp;
end;

{***************************************************************}
procedure TBimAppCmd.OnAppCmdReconTimer(Sender: TObject);
begin
  FTimerRecon.Enabled := False;
  TryToLogin;
end;

{***************************************************************}
procedure TBimAppCmd.OnAppCmdSendTimer(Sender: TObject);
var sAcc: string;
begin
  if (FListSend.Count = 0) then
  begin
    FTimerSend.Enabled := False;
    if (FSendCount > 0) then
    begin
      Log('* Done! Messages sent: ' + IntToStr(FSendCount));
      FSendCount := 0;
    end;

    if not FCmdParams.OtherInfo.StayConnected then
      QuitApp;

    Exit;
  end;

  if FConnected then
  begin
    sAcc := StringReplace(Trim(FListSend.Strings[0]), SPC, '', [rfReplaceAll]);
    FListSend.Delete(0);

    SendMsgToAcc(sAcc);
  end;

  //quick finish and report
  if (FListSend.Count = 0) then
  begin
    FTimerSend.Enabled  := False;
    FTimerSend.Interval := 250;
    FTimerSend.Enabled  := True;
  end;
end;

{***************************************************************}
procedure TBimAppCmd.SendMsgToAcc(Acc: string);
var SendMsgInfo: TSendMsgInfo;
    iRes: integer;
begin
  with SendMsgInfo do
  begin
    Account       := Acc;
    MsgUniqID     := GetUniqSequence;
    MsgText       := FCmdParams.MsgInfo.MsgText;
    AckRequired   := FCmdParams.MsgInfo.MsgAck;
    TranspID      := 0; //must be 0, this is for future server versions
  end;//with

  //send message
  iRes := ObimpProto.SendMsg(@SendMsgInfo);
  if (iRes = FUNC_OPER_SUCCESS) then
  begin
    Inc(FSendCount);
    Log('* Message has been sent to "' + Acc + '"');
  end
  else
    Log('! Can''t send message to "' + Acc + '". Some parameter is wrong.');
end;

{ TObimpEvents }
{***************************************************************}
procedure TObimpEvents.AuthReplyRcvd(AuthReply: pRcvdAuthReply);
var s: string;
begin
  s := '';
  case AuthReply^.ReplyCode of
    AUTH_REPLY_GRANTED : s := 'granted';
    AUTH_REPLY_DENIED  : s := 'denied';
  end;//case

  Log('# Event: Authorization reply "' + s + '" from "' + AuthReply^.Account + '"');
end;

{***************************************************************}
procedure TObimpEvents.AuthReqRcvd(AuthReq: pRcvdAuthReq);
var AuthReply: TSendAuthReply;
    iRes: integer;
    s: string;
begin
  if (Trim(AuthReq^.ReasonText) = '') then
    Log('# Event: Authorization request from "' + AuthReq^.Account + '"')
  else
    Log('# Event: Authorization request from "' + AuthReq^.Account + '": ' + AuthReq^.ReasonText);

  //check if we need to reply
  s := '';
  case AppCmd.FCmdParams.OtherInfo.AuthReply of
    AUTH_REPLY_GRANTED : s := 'granted';
    AUTH_REPLY_DENIED  : s := 'denied';
  end;//case

  if (s = '') then
    Exit;

  //grant or deny authorization request using:
  with AuthReply do
  begin
    Account    := AuthReq^.Account;
    ReplyCode  := AppCmd.FCmdParams.OtherInfo.AuthReply;
    TranspID   := 0; //must be 0, this is for future server versions
  end;//with

  iRes := AppCmd.ObimpProto.SendAuthReply(@AuthReply);
  if (iRes = FUNC_OPER_SUCCESS) then
    Log('* Authorization reply (' + s + ') has been sent to "' + AuthReq^.Account + '"')
  else
    Log('! Failed to send authorization reply to "' + AuthReq^.Account + '"');
end;

{***************************************************************}
procedure TObimpEvents.AuthRevokeRcvd(AuthRevoke: pRcvdAuthRevoke);
begin
  if (Trim(AuthRevoke^.ReasonText) = '') then
    Log('# Event: Authorization revoke from "' + AuthRevoke^.Account + '"')
  else
    Log('# Event: Authorization revoke from "' + AuthRevoke^.Account + '": ' + AuthRevoke^.ReasonText);
end;

{***************************************************************}
procedure TObimpEvents.ConnectionFailed;
begin
  Log('# Event: Connection failed.');
  AppCmd.FConnected := False;
  AppCmd.ReconnectRequired;
end;

{***************************************************************}
function TObimpEvents.GetProtoIntfVer: DWord;
begin
  //return SDK protocol interface version like it done below
  Result := PROTO_INTF_VERSION;
end;

{***************************************************************}
procedure TObimpEvents.LoginFailed(LoginFailedCode: Word);
  function LoginFailedDescr: string;
  begin
    Result := '';
    case LoginFailedCode of
      LOGIN_FAILED_INVALID_LOGIN    : Result := 'LOGIN_FAILED_INVALID_LOGIN';
      LOGIN_FAILED_TEMP_UNAVAILABLE : Result := 'LOGIN_FAILED_TEMP_UNAVAILABLE';
      LOGIN_FAILED_TOO_MANY_CLIENTS : Result := 'LOGIN_FAILED_TOO_MANY_CLIENTS';
      LOGIN_FAILED_OTHER_SRV_ERROR  : Result := 'LOGIN_FAILED_OTHER_SRV_ERROR';
    end;//case
  end;
begin
  Log('# Event: Login failed (code=' + LoginFailedDescr + ').');
  AppCmd.QuitApp;
end;

{***************************************************************}
procedure TObimpEvents.LoginSuccess;
begin
  Log('# Event: Login done.');
  AppCmd.FConnected := True;
  AppCmd.ReadyToSendMsg;
end;

{***************************************************************}
procedure TObimpEvents.MsgAckRcvd(MsgAck: pMsgAck);
begin
  Log('# Event: Your message delivered to "' + MsgAck^.Account + '"');
end;

{***************************************************************}
procedure TObimpEvents.MsgReceived(RcvdMsgInfo: pRcvdMsgInfo);
var MsgAck: TMsgAck;
    iRes: integer;
begin
  Log('# Event: New message from "' + RcvdMsgInfo^.Account + '": ' + Copy(RcvdMsgInfo^.MsgText, 1, 16384));

  //If sender requested message acknowledge then we should send it
  //(ofcourse if we are not in invisible statuses, act here like you wish).
  if RcvdMsgInfo^.AckRequired then
  begin
    with MsgAck do
    begin
      Account   := RcvdMsgInfo^.Account;
      MsgUniqID := RcvdMsgInfo^.MsgUniqID;
      TranspID  := RcvdMsgInfo^.TranspID;
    end;//with

    iRes := AppCmd.ObimpProto.SendMsgAck(@MsgAck);
    if (iRes = FUNC_OPER_SUCCESS) then
      Log('* Sent delivery report to "' + RcvdMsgInfo^.Account + '"')
    else
      Log('! Failed to send delivery report to "' + RcvdMsgInfo^.Account + '"');
  end;
end;

{***************************************************************}
procedure TObimpEvents.ServerBye(ByeReasonCode: Word);
  function ServerByeDescr: string;
  begin
    Result := '';
    case ByeReasonCode of
      BYE_REASON_SRV_SHUTDOWN       : Result := 'BYE_REASON_SRV_SHUTDOWN';
      BYE_REASON_CLI_NEW_LOGIN      : Result := 'BYE_REASON_CLI_NEW_LOGIN';
      BYE_REASON_ACCOUNT_KICKED     : Result := 'BYE_REASON_ACCOUNT_KICKED';
      BYE_REASON_INCORRECT_SEQ      : Result := 'BYE_REASON_INCORRECT_SEQ';
      BYE_REASON_INCORRECT_BEX_TYPE : Result := 'BYE_REASON_INCORRECT_BEX_TYPE';
      BYE_REASON_INCORRECT_BEX_SUB  : Result := 'BYE_REASON_INCORRECT_BEX_SUB';
      BYE_REASON_INCORRECT_BEX_STEP : Result := 'BYE_REASON_INCORRECT_BEX_STEP';
      BYE_REASON_TIMEOUT            : Result := 'BYE_REASON_TIMEOUT';
      BYE_REASON_INCORRECT_WTLD     : Result := 'BYE_REASON_INCORRECT_WTLD';
      BYE_REASON_NOT_ALLOWED        : Result := 'BYE_REASON_NOT_ALLOWED';
      BYE_REASON_FLOODING           : Result := 'BYE_REASON_FLOODING';
    end;//case
  end;
begin
  Log('# Event: ServerBye (code=' + ServerByeDescr + ').');

  if (ByeReasonCode = BYE_REASON_CLI_NEW_LOGIN) then
    AppCmd.QuitApp
  else
  begin
    AppCmd.FConnected := False;
    AppCmd.ReconnectRequired;
  end;
end;

end.
